"""
Absufyu: Checksum
-----------------
Check MD5, SHA256, ...

Version: 5.17.0
Date updated: 20/11/2025 (dd/mm/yyyy)
"""

# Module level
# ---------------------------------------------------------------------------
__all__ = [
    # Checksum
    "Checksum",
    "ChecksumMode",
    # Mixin
    "DirectoryRemoveDuplicateMixin",
]


# Library
# ---------------------------------------------------------------------------
import hashlib
from enum import StrEnum
from pathlib import Path
from typing import Literal

from absfuyu.core.baseclass import BaseClass
from absfuyu.core.docstring import deprecated, versionadded, versionchanged
from absfuyu.core.dummy_func import tqdm
from absfuyu.dxt import DictExt, ListExt
from absfuyu.util.path import DirectoryBase


# Function
# ---------------------------------------------------------------------------
@deprecated("5.0.0")
def _checksum_operation(
    file: Path | str,
    hash_mode: str | Literal["md5", "sha1", "sha256", "sha512"] = "sha256",
) -> str:
    """
    This performs checksum
    """
    if hash_mode.lower() == "md5":
        hash_engine = hashlib.md5()
    elif hash_mode.lower() == "sha1":
        hash_engine = hashlib.sha1()
    elif hash_mode.lower() == "sha256":
        hash_engine = hashlib.sha256()
    elif hash_mode.lower() == "sha512":
        hash_engine = hashlib.sha512()
    else:
        hash_engine = hashlib.md5()

    with open(Path(file), "rb") as f:
        while True:
            data = f.read(4096)
            if len(data) == 0:
                break
            else:
                hash_engine.update(data)
    return hash_engine.hexdigest()


# Class
# ---------------------------------------------------------------------------
class ChecksumMode(StrEnum):
    MD5 = "md5"
    SHA1 = "sha1"
    SHA256 = "sha256"
    SHA512 = "sha512"


class DuplicateSummary(DictExt[str, list[Path]]):
    """
    Duplicate file summary
    """

    def summary(self) -> int:
        """
        Show how many duplicates (include the original)

        Returns
        -------
        int
            How many duplicates
        """
        temp = self.__class__(self.copy())
        try:
            return sum(temp.apply(lambda x: len(x)).values())
        except Exception as err:
            print(f"Something wrong - {err}")

    def remove_duplicates(self, dry_run: bool = True, keep_first: bool = True, debug: bool = False) -> None:
        """
        Remove duplicates

        Parameters
        ----------
        dry_run : bool, optional
            Simulate only (no files deleted), by default ``True``

        keep_first : bool, optional
            Keep the first duplicate file, will keep the last duplicate file when ``False``, by default ``True``
        """
        temp = self.__class__(self.copy())
        removable_files = ListExt([x[1:] if keep_first else x[:-1] for x in temp.values()]).flatten()

        for x in removable_files:
            x: Path = x

            if debug or dry_run:
                print(f"Deleting {x}")
            if dry_run:
                continue

            x.unlink(missing_ok=True)


@versionchanged("4.1.1", reason="Checksum for entire folder is possible")
@versionadded("4.1.0")
class Checksum(BaseClass):
    """
    Checksum engine

    Parameters
    ----------
    path : str | Path
        Path to file/directory to perform checksum

    hash_mode : ChecksumMode | Literal["md5", "sha1", "sha256", "sha512"], optional
        Hash mode, by default ``"sha256"``

    save_result_to_file : bool, optional
        Save checksum result(s) to file, by default ``False``
    """

    def __init__(
        self,
        path: str | Path,
        hash_mode: (
            ChecksumMode | Literal["md5", "sha1", "sha256", "sha512"]
        ) = ChecksumMode.SHA256,
        save_result_to_file: bool = False,
    ) -> None:
        """
        Checksum engine

        Parameters
        ----------
        path : str | Path
            Path to file/directory to perform checksum

        hash_mode : ChecksumMode | Literal["md5", "sha1", "sha256", "sha512"], optional
            Hash mode, by default ``"sha256"``

        save_result_to_file : bool, optional
            Save checksum result(s) to file, by default ``False``
        """
        self.path = Path(path)
        self.hash_mode = hash_mode
        self.save_result_to_file = save_result_to_file
        self.checksum_result_file_name = "checksum_results.txt"

    def _get_hash_engine(self):
        hash_mode = self.hash_mode
        if hash_mode.lower() == "md5":
            hash_engine = hashlib.md5()
        elif hash_mode.lower() == "sha1":
            hash_engine = hashlib.sha1()
        elif hash_mode.lower() == "sha256":
            hash_engine = hashlib.sha256()
        elif hash_mode.lower() == "sha512":
            hash_engine = hashlib.sha512()
        else:
            hash_engine = hashlib.md5()
        return hash_engine

    def _checksum_operation(
        self,
        file: Path | str,
    ) -> str:
        """This performs checksum"""

        hash_engine = self._get_hash_engine().copy()
        # with open(Path(file), "rb") as f:
        with file.open("rb") as f:
            # Read and hash the file in 4K chunks. Reading the whole
            # file at once might consume a lot of memory if it is
            # large.
            while True:
                data = f.read(4096)
                if len(data) == 0:
                    break
                else:
                    hash_engine.update(data)
        return hash_engine.hexdigest()  # type: ignore

    def checksum(self, recursive: bool = True) -> str:
        """
        Perform checksum

        Parameters
        ----------
        recursive : bool, optional
            Do checksum for every file in the folder (including child folder),
            by default ``True``

        Returns
        -------
        str
            Checksum hash
        """
        if self.path.absolute().is_dir():  # Dir
            new_path = self.path.joinpath(self.checksum_result_file_name)
            # List of files
            if recursive:
                file_list: list[Path] = [
                    x for x in self.path.glob("**/*") if x.is_file()
                ]
            else:
                file_list = [x for x in self.path.glob("*") if x.is_file()]

            # Checksum
            res = []
            for x in tqdm(file_list, desc="Calculating hash", unit_scale=True):
                name = x.relative_to(self.path)
                res.append(f"{self._checksum_operation(x)} | {name}")
            output = "\n".join(res)
        else:  # File
            new_path = self.path.with_name(self.checksum_result_file_name)
            output = self._checksum_operation(self.path)

        # Save result
        if self.save_result_to_file:
            with open(new_path, "w", encoding="utf-8") as f:
                f.write(output)

        return output


# Mixin
class DirectoryRemoveDuplicateMixin(DirectoryBase):
    """
    Directory - Remove duplicate by SHA256

    - remove_duplicate
    """

    def __init__(self, source_path, create_if_not_exist=False) -> None:
        super().__init__(source_path, create_if_not_exist)

        self._duplicate_cache: DuplicateSummary | None = None

    def _gather_duplicate_cache(self, recursive: bool = True) -> None:
        engine = Checksum(self.source_path, hash_mode=ChecksumMode.SHA256, save_result_to_file=False)
        valid = [x for x in engine.path.glob("**/*" if recursive else "*") if x.is_file()]
        checksum_cache = {}

        # Checksum
        for x in tqdm(valid, unit_scale=True, desc="Checking..."):
            try:
                cs_res = engine._checksum_operation(x)

                if checksum_cache.get(cs_res) is None:
                    checksum_cache[cs_res] = [x]
                else:
                    checksum_cache[cs_res] += [x]
            except Exception as err:
                print(f"ERROR: {x} - {err}")
                continue

        # Save to cache
        self._duplicate_cache = DuplicateSummary({k: v for k, v in checksum_cache.items() if len(v) > 1})

    def remove_duplicate(self, dry_run: bool = True, recursive: bool = True, debug: bool = True) -> None:
        """
        Remove duplicate files by SHA256 checksum

        Parameters
        ----------
        dry_run : bool, optional
            Simulate only (no files deleted), by default ``True``

        recursive : bool, optional
            Scan every file in the folder (including child folder), by default ``True``

        debug : bool, optional
            Print delete messages, by default ``True``
        """
        self._gather_duplicate_cache(recursive=recursive)

        # Remove
        try:
            summary = self._duplicate_cache
            print(f"Duplicate files: {summary.summary()}")
            summary.remove_duplicates(dry_run=dry_run, keep_first=False, debug=debug)
        except Exception as err:
            pass



"""
Absfuyu: PDF
------------
PDF Tool [W.I.P]

Version: 5.17.0
Date updated: 20/11/2025 (dd/mm/yyyy)
"""

# Module level
# ---------------------------------------------------------------------------
__all__ = ["PDFTool"]


# Library
# ---------------------------------------------------------------------------
from collections.abc import Sequence
from pathlib import Path
from typing import overload

from absfuyu.core import BaseClass
from absfuyu.core.dummy_func import tqdm

PDF_MODE = False

try:
    from spire.pdf import PdfDocument, PdfTextExtractOptions, PdfTextExtractor
except ImportError:
    from subprocess import run

    from absfuyu.config import ABSFUYU_CONFIG

    if ABSFUYU_CONFIG._get_setting("auto-install-extra").value:  # type: ignore
        cmd = "python -m pip install -U absfuyu[full]".split()
        run(cmd)
    else:
        raise SystemExit("This feature is in absfuyu[full] package")  # noqa: B904
else:
    PDF_MODE = True


# Class
# ---------------------------------------------------------------------------
class PDFTool(BaseClass):
    def __init__(self) -> None:
        super().__init__()
        self.engine = PdfDocument()  # type: ignore

        self.files: list[Path] = []

    @overload
    def add_file(self, file: Path | str) -> None: ...

    @overload
    def add_file(self, file: Sequence[Path | str]) -> None: ...

    def add_file(self, file: Path | Sequence[Path | str] | str) -> None:
        if isinstance(file, Sequence) and not isinstance(file, str):
            self.files.extend([Path(x) for x in file])
        else:
            self.files.append(Path(file))

    def load_file(self) -> list[str]:
        """
        Extract text from all file

        Returns
        -------
        list[str]
            Extracted text
        """
        engine = PdfDocument()  # type: ignore

        extracted: list[str] = []
        for x in tqdm(self.files, desc="Extracting", unit_scale=True):
            engine.LoadFromFile(x.absolute().__str__())

            # Get pages
            for page in engine.Pages:
                # Create a PdfTextExtractor for the page
                extractor = PdfTextExtractor(page)  # type: ignore

                # Extract all text from the page
                text = extractor.ExtractText(PdfTextExtractOptions())  # type: ignore
                extracted.append(text)

        return extracted



"""
Absufyu: Utilities
------------------
Some random utilities

Version: 5.17.0
Date updated: 20/11/2025 (dd/mm/yyyy)
"""

# Module Package
# ---------------------------------------------------------------------------
__all__ = [
    # util
    "get_installed_package",
    "set_min",
    "set_max",
    "set_min_max",
    "stop_after_day",
    "convert_to_raw_unicode",
    # shorten number
    "Decimal",
]


# Library
# ---------------------------------------------------------------------------
import pkgutil
import subprocess
from datetime import datetime
from string import printable

from absfuyu.core.docstring import deprecated, versionadded, versionchanged
from absfuyu.util.shorten_number import Decimal


# Function
# ---------------------------------------------------------------------------
@versionchanged("2.7.1", reason="Use ``pkgutil`` lib")
def get_installed_package():
    """
    Return a list of installed packages

    Returns
    -------
    list[str]
        List of installed packages
    """
    iter_modules = list(
        {module.name for module in pkgutil.iter_modules() if module.ispkg}
    )
    return sorted(iter_modules)


@deprecated("5.0.0")
def set_min(
    current_value: int | float,
    *,
    min_value: int | float = 0,
) -> int | float:
    """
    Return ``min_value`` when ``current_value`` < ``min_value``

    Parameters
    ----------
    current_value : int | float
        Current value

    min_value : int | float
        Minimum value
        (Default: ``0``)

    Returns
    -------
    int | float
        Analyzed value


    Example:
    --------
    >>> set_min(-1)
    0
    """
    # if current_value < min_value:
    #     current_value = min_value
    # return current_value
    return max(min_value, current_value)


@deprecated("5.0.0")
def set_max(
    current_value: int | float,
    *,
    max_value: int | float = 100,
) -> int | float:
    """
    Return ``max_value`` when ``current_value`` > ``max_value``

    Parameters
    ----------
    current_value : int | float
        Current value

    max_value : int | float
        Maximum value
        (Default: ``100``)

    Returns
    -------
    int | float
        Analyzed value


    Example:
    --------
    >>> set_max(101)
    100
    """
    # if current_value > max_value:
    #     current_value = max_value
    # return current_value
    return min(max_value, current_value)


def set_min_max(
    current_value: int | float,
    *,
    min_value: int | float = 0,
    max_value: int | float = 100,
) -> int | float:
    """
    Return ``min_value`` | ``max_value`` when ``current_value``
    is outside ``[min_value, max_value]``

    Parameters
    ----------
    current_value : int | float
        Current value

    min_value : int | float
        Minimum value
        (Default: ``0``)

    max_value : int | float
        Maximum value
        (Default: ``100``)

    Returns
    -------
    int | float
        Analyzed value


    Example:
    --------
    >>> set_min_max(808)
    100
    """
    # Set min
    # current_value = set_min(current_value, min_value=min_value)
    current_value = max(current_value, min_value)
    # Set max
    # current_value = set_max(current_value, max_value=max_value)
    current_value = min(current_value, max_value)
    return current_value


@versionchanged("5.6.0", reason="New `custom_msg` parameter")
@versionadded("3.2.0")
def stop_after_day(
    year: int | None = None,
    month: int | None = None,
    day: int | None = None,
    *,
    custom_msg: str | None = None,
) -> None:
    """
    Stop working after specified day.
    Put the function at the begining of the code is recommended.

    Parameters
    ----------
    year : int
        Desired year
        (Default: ``None``)

    month : int
        Desired month
        (Default: ``None``)

    day : int
        Desired day
        (Default: ``None`` - 1 day trial)

    custom_msg : str
        Custom exit message
        (Default: ``None``)
    """
    # None checking - By default: 1 day trial
    now = datetime.now()
    if year is None:
        year = now.year
    if month is None:
        month = now.month
    if day is None:
        day = now.day + 1

    # Logic
    end_date = datetime(year, month, day)
    result = end_date - now
    if result.days < 0:
        if custom_msg:
            raise SystemExit(custom_msg)
        raise SystemExit("End of time")


@versionadded("5.4.0")
def convert_to_raw_unicode(text: str, partial: bool = True) -> str:
    r"""
    Convert text to raw unicode variant.

    Parameters
    ----------
    text : str
        Text to convert

    partial : bool, optional
        Only convert characters that not in ``string.printable``,
        by default ``True``

    Returns
    -------
    str
        Converted text.


    Example:
    --------
    >>> convert_to_raw_unicode("résumé")
    r\u00E9sum\u00E9

    >>> convert_to_raw_unicode("résumé", partial=False)
    \u0072\u00E9\u0073\u0075\u006D\u00E9
    """

    character_set = printable

    def _convert(character: str) -> str:
        """Get unicode value"""
        # ord(c): Returns the Unicode code point for a one-character string c.
        _ord = ord(character)
        # f"\\u{_ord:04X}": Formats the Unicode code point as a four-digit
        # hexadecimal number for code points less than 0x10000.
        # f"\\U{_ord:08X}": Formats the Unicode code point as an eight-digit
        # hexadecimal number for code points greater than or equal to 0x10000.
        return f"\\u{_ord:04X}" if _ord < 0x10000 else f"\\U{_ord:08X}"

    return "".join(
        character if character in character_set and partial else _convert(character)
        for character in text
    )


@versionadded("5.9.0")
def is_command_available(cmd: list[str] | str, err_msg: str = "") -> None:
    """
    Checks if the desired command available

    Parameters
    ----------
    cmd : list[str] | str
        Command to check

    err_msg : str, optional
        Custom error message, by default ""

    Raises
    ------
    ValueError
        When command is unvailable
    """
    try:
        subprocess.run(
            cmd,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
    except Exception:
        raise ValueError(err_msg)
